#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
using namespace std;

#define MaxN 50
#define SIMTIME  1000000
#define QUEUESIZE 500 // used to create arrays
#define MaxSQRTC 6



typedef struct queuestruct {
    double q[QUEUESIZE][2];
    int si; // start index
    int count; // count
}queuetype;


int V;
int N;
int SQRTCX;
int SQRTCY;
int x[MaxN];
int y[MaxN];
double cell[MaxSQRTC][MaxSQRTC][6];
int U[MaxN][MaxN];
double t;
FILE *fout;
queuetype quar[MaxN][MaxN][2];
double delay;
double delaysum;
double difdelay;
int qsum;

double rand_p;
double lambda;
int timeout;
double sum_num_transmissions;
double receivedcount;
double pcktscount;
double seq_num[MaxN];
double seq_rcvd[MaxN];
int pcktrcvd[500000];
int maxdupQsize;
double dupQempty;
double bothQempty;
int initializeq(queuetype *qp)
{
    qp->si = 0;
    qp->count = 0;
    return (0);
}

int pushq(queuetype *qp, double v1, double v2)
{
    if (qp->count >= QUEUESIZE - 2) {      // if full
       fprintf(fout, "ERROR (push) queue is full\n");
       exit(1);
       
    }
    qp->q[(qp->si + qp->count) % QUEUESIZE][0] = v1;
    qp->q[(qp->si + qp->count) % QUEUESIZE][1] = v2;
    qp->count++;
    return(0);
}

double popq(queuetype *qp, double *v2)
{
    if (qp->count <= 0) { //if empty
       fprintf(fout, "ERROR (pop) queue is empty\n");
       exit(1);
    }
    double v1;
    v1 = qp->q[qp->si][0];
    *v2 = qp->q[qp->si][1];
    qp->si++;
    qp->si %=  QUEUESIZE;
    qp->count--;
    return(v1);    
}

double observeq(queuetype *qp, double *v2) // read head without pop
{
    if (qp->count <= 0) { //if empty
		return(-1);
    }
    double v1;
    v1 = qp->q[qp->si][0];
    *v2 = qp->q[qp->si][1];
    return(v1);    
}


double poplifoq(queuetype *qp, double *v2)
{
    if (qp->count <= 0) { //if empty
       fprintf(fout, "ERROR (pop) queue is empty\n");
       exit(1);
    }
    double v1;
	
    v1 = qp->q[(qp->si + qp->count - 1) % QUEUESIZE][0];
    *v2 = qp->q[(qp->si + qp->count - 1) % QUEUESIZE][1];
    qp->count--;
    return(v1);    
}


int isemptyq(queuetype *qp)
{
    if (qp->count <= 0) { //if empty
       return(1);
    }
    return (0);
}

int isfullq(queuetype *qp)
{
    if (qp->count >= QUEUESIZE - 2) {      // if full
       return(1);
    }
    return (0);
}



int uniform_rnd(int n)
{
	int r;
	r = (int) (((double)rand() / ((double)(RAND_MAX) + (double)(1))) * n);
	return (r);
}


/* return uniform random variable in [0,1) */
double uniform_rnd_float()
{
	return(  (double)rand()  / (    (double)(RAND_MAX) + (double)(1)   ) );
}

void init_pos()
{
	int i;
	for(i = 0; i < N; i++){
		x[i] = uniform_rnd(SQRTCX);
		y[i] = uniform_rnd(SQRTCY);
	}
}

int is_received(double v1)
{
    int i;
    int j;
    int n;
    i = int(v1) / (int) 16;
    j = int(v1) % (int) 16;
    n = 1;
    n = n << j;
    n = n & pcktrcvd[i];
    if(n == 0) {
         return(0);
    }
    return(1);
}

int packet_received(double v1, double v2)
{
    int i;
    int j;
    int n;
	// v1 packet seqnum
	// v2 packet start time
	
	// if not already received
	if(!(is_received(v1))) {
    
         // update average delay
	     delaysum += t - v2;
	     receivedcount++;
	
	     // register this packet as already received
	     i = int(v1) / (int) 16;
	     j = int(v1) % (int) 16;
	     n = 1;
	     n = n << j;
	     pcktrcvd[i] = pcktrcvd[i] | n;
    }
     
	// else 
	// do nothing
	return(0);
}

int SIM()
{
	/*
	 * set to zero
	 */
	
	delaysum = 0;
    delay = 0;
	
	sum_num_transmissions = 0;
    receivedcount = 0;
	pcktscount = 0;
	int i;
	int ii;
	int j;
	int k;
	int tr;
	int rc;
	int cm;
	int cp;
	int cpflag;
	double v1; // packet seqnum
	double v2; // packet start time
	double v1temp; // packet seqnum
	double v2temp; // packet start time
	
	double wabc;
	double wabct;
	
	double tiecount;
	
	for(i = 0; i < N; i++){
        for(j = 0; j < N; j++){  
            initializeq(&(quar[i][j][0]));
			initializeq(&(quar[i][j][1]));
            U[i][j] = 0;
        }
	}
	for(i = 0; i < 500000; i++) {
          pcktrcvd[i] = 0;
    }
	init_pos();
	t = 0;

	for(i = 0; i < N; i++) {
          seq_num[i] = 0;
          seq_rcvd[i] = 0;
    }
    maxdupQsize = 0;
	dupQempty = 0;
	bothQempty = 0;    
	while (t < SIMTIME) {
		init_pos();
		
		//duplicates removal
		for(i = 0; i < N; i++) {
			  for(j = 0; j < N; j++) {
				  if ( i !=j) {
					    for(ii = 0; ii < quar[i][j][1].count; ii++) {
								U[i][j]--;
								v1temp = popq(&(quar[i][j][1]), &v2temp);
								//if(!is_received(j * lambda * 1.2 * SIMTIME + v1temp)){
								if (v2temp < 0){
									//removing original packets in the dublicate buffer if already received
									if((x[i] == x[j]) && (y[i] == y[j])) {
											if(!is_received(j * lambda * 1.2 * SIMTIME + v1temp)){
												if(t + v2temp < timeout ) { // v2temp < 0
													U[i][j]++;
													pushq(&(quar[i][j][1]),v1temp, v2temp);
												}
												else {
													U[i][j]++;
													v2temp = - v2temp;
													pushq(&(quar[i][j][0]),v1temp, v2temp);
												}
											}
									}
									else {
										U[i][j]++;
										pushq(&(quar[i][j][1]),v1temp, v2temp);
									}
								}
								else if (t - v2temp < timeout) {
									U[i][j]++;
									pushq(&(quar[i][j][1]),v1temp, v2temp);
								}
						}
				  }
			  }
		}
		
		//initialize cell
		for(i = 0; i < SQRTCX; i++) {
              for(j = 0; j < SQRTCY; j++) {
					cell[i][j][4] = 0;
					cell[i][j][5] = 0;
					cell[i][j][0] = -1;
					cell[i][j][1] = -2;
					cell[i][j][2] = -3;
					cell[i][j][3] = 0;  //do not copy
              }
        }
		// computing m_ab, m_ab^c(t)
		tiecount = 1.0;
		for(i = 0; i < N; i++) {
              for(j = 0; j < N; j++) {
                    if (i != j) {
						  if(quar[i][j][1].count > maxdupQsize) {
								maxdupQsize = quar[i][k][1].count;
						  }
						  if(quar[i][j][1].count == 0){
								dupQempty++;
								if(quar[i][j][0].count == 0) {
									bothQempty++;
								}
						  }

                          if((x[i] == x[j]) && (y[i] == y[j])) {

							    // node i & j in the same cell 
                                 for(k = 0; k < N; k++){
                                       // for each comodity k
                                       wabc =  quar[i][k][0].count - quar[j][k][0].count; 
									   wabct = U[i][k] - U[j][k]; //quar[j][k][0].count; // -+ 0.005 * quar[j][k][1].count ;
									   if(U[i][k] == 0){
										   // do nothing
									   }
									   else {
                                       		if(wabc > cell[x[i]][y[i]][4]) {
											   tiecount = 1.0;
                                               cell[x[i]][y[i]][4] = wabc;
											   cell[x[i]][y[i]][5] = wabct;
                                               cell[x[i]][y[i]][0] = i;
                                               cell[x[i]][y[i]][1] = j;
                                               cell[x[i]][y[i]][2] = k;
                                               cell[x[i]][y[i]][3] = 0;  //do not copy
											   if(U[i][k] <= V){
												   cell[x[i]][y[i]][3] = 1;  //copy
											   }
                                       		}
									   		else if(wabc == cell[x[i]][y[i]][4]) {
										   		if (j == k) {
												   if(cell[x[i]][y[i]][1] != cell[x[i]][y[i]][2]) {
														tiecount = 1.0;
													    cell[x[i]][y[i]][4] = wabc;
														cell[x[i]][y[i]][5] = wabct;
														cell[x[i]][y[i]][0] = i;
														cell[x[i]][y[i]][1] = j;
														cell[x[i]][y[i]][2] = k;
														cell[x[i]][y[i]][3] = 0;  //do not copy
														if(U[i][k] <= V){
															cell[x[i]][y[i]][3] = 1;  //copy
														}
												   }
												   else if (wabct > cell[x[i]][y[i]][5]) {
														tiecount = 1.0;
													    cell[x[i]][y[i]][4] = wabc;
														cell[x[i]][y[i]][5] = wabct;
														cell[x[i]][y[i]][0] = i;
														cell[x[i]][y[i]][1] = j;
														cell[x[i]][y[i]][2] = k;
														cell[x[i]][y[i]][3] = 0;  //do not copy
														if(U[i][k] <= V){
															cell[x[i]][y[i]][3] = 1;  //copy
														}

												   }
												   else if (wabct == cell[x[i]][y[i]][5]) {
													    tiecount++;
														rand_p = uniform_rnd_float();
														if(rand_p < 1.0 / tiecount) {
															cell[x[i]][y[i]][4] = wabc;
															cell[x[i]][y[i]][5] = wabct;
															cell[x[i]][y[i]][0] = i;
															cell[x[i]][y[i]][1] = j;
															cell[x[i]][y[i]][2] = k;
															cell[x[i]][y[i]][3] = 0;  //do not copy
															if(U[i][k] <= V){
																cell[x[i]][y[i]][3] = 1;  //copy
															}
														}
												   }
												}
										   		else if(cell[x[i]][y[i]][1] != cell[x[i]][y[i]][2]) {
											   		if (wabct > cell[x[i]][y[i]][5]) {
														tiecount = 1.0;
													    cell[x[i]][y[i]][4] = wabc;
														cell[x[i]][y[i]][5] = wabct;
														cell[x[i]][y[i]][0] = i;
														cell[x[i]][y[i]][1] = j;
														cell[x[i]][y[i]][2] = k;
														cell[x[i]][y[i]][3] = 0;  //do not copy
														if(U[i][k] <= V){
															cell[x[i]][y[i]][3] = 1;  //copy
														}
											   		}
											   		else if (wabct == cell[x[i]][y[i]][5]) {
													    tiecount++;
														rand_p = uniform_rnd_float();
														if(rand_p < 1.0 / tiecount) {
															cell[x[i]][y[i]][4] = wabc;
															cell[x[i]][y[i]][5] = wabct;
															cell[x[i]][y[i]][0] = i;
															cell[x[i]][y[i]][1] = j;
															cell[x[i]][y[i]][2] = k;
															cell[x[i]][y[i]][3] = 0;  //do not copy
															if(U[i][k] <= V){
																cell[x[i]][y[i]][3] = 1;  //copy
															}
														}
											   		}
										   		}
									   		}
									   }
                                 }
                          }
                    }
              }
        }
        // updating queues
		for(i = 0; i < SQRTCX; i++) {
              for(j = 0; j < SQRTCY; j++) {
                    if ( (cell[i][j][4] > 0) || (cell[i][j][5] > 0)) {
                         //we have packet transmition 
                         // from node cell[i][j][0]
                         tr = cell[i][j][0];
                         // to node cell[i][j][1]
                         rc = cell[i][j][1];
                         // comodity cell[i][j][2]
                         cm = cell[i][j][2];
                         //copy or not cell[i][j][3]    NOT USED
                         cp = cell[i][j][3];   // not used

                         
						 if(!(isemptyq(&(quar[tr][cm][0])))) {
							v1 = popq(&(quar[tr][cm][0]), &v2);
							U[tr][cm]--;
							cpflag = 0;  // indication that this packet from original queue
						 }
						 else {
							v1 = popq(&(quar[tr][cm][1]), &v2); // poplifoq
							U[tr][cm]--;
							cpflag = 1;  // indication that this packet from duplicates queue
						 }
						 if(rc == cm){
                             // packet received
							 sum_num_transmissions++;
							 if(v2 < 0) {
								 v2 = - v2;
							 }
                             packet_received(cm * lambda * 1.2 * SIMTIME + v1, v2);
                             if(v1 == seq_rcvd[cm]){
                                   seq_rcvd[cm]++;
                                   while(is_received(cm * lambda * 1.2 * SIMTIME + seq_rcvd[cm])) {
                                       seq_rcvd[cm]++;    
                                   }        
                             }
                         }
						 else {
							  int origflag = 0;
							  if(U[tr][cm] < V) {
									if (v2 < 0){
										// when v2 < 0 then this packet is an original packet but kept in the dublicate buffer
										// since we are here creating a dublicate we make sure v2 > 0
										v2 = - v2;
										origflag = 1;
									}
									pushq(&(quar[tr][cm][1]),v1, v2);
									U[tr][cm]++;
									if(cpflag == 0){
										cpflag = 1;
										v2 = - v2; // original packet to be placed in dublicate buffer we make sure v2 < 0
									}
							  }
							  if(origflag ==1) {
								  v2 = -v2; // return v2 back to be less than zero
							  }
                              sum_num_transmissions++;
							  pushq(&(quar[rc][cm][cpflag]),v1, v2);
							  U[rc][cm]++;                              
                         }
					}
              }
        }
        
        // arrivals 
		for(i = 0; i < N; i++) {
              rand_p = uniform_rnd_float();
			  if(rand_p < lambda) {
				  // we have an arrival
				  pcktscount++;
				  if (i % 2 == 0) {
					  j = i + 1;
				  }
				  else {
					  j = i - 1;
				  }
                  U[i][j]++;
                  pushq(&(quar[i][j][0]),seq_num[j], t);
                  (seq_num[j])++;
              }        
        }
		t++;
	}
	qsum = 0;
	for(i = 0; i < N; i++) {
		for(j = 0; j < N; j++) {
			qsum += U[i][j];
		}
	}


	difdelay = delay - (delaysum / receivedcount);
    delay = delaysum / receivedcount;
	return(0);
}





int main(int argc, char *argv[])
{
    srand((unsigned)time(NULL));
    fout = fopen("out.txt","wt");
    lambda = 0.128;
	timeout = 1; 
    while (timeout <= 2048) {
		  SQRTCX = 5;
          while (  SQRTCX <= 5) {
			    SQRTCY = SQRTCX;
			    N = (int) SQRTCX * SQRTCY * 1.7399;
				if (N % 2 == 1) {
					N++;
				}
                V = 1;
                while (V <= 1) {
                      SIM();
                      fprintf(fout,"lambda = %f\n", lambda);
                      fprintf(fout,"N = %d\n", N);
                      fprintf(fout,"V = %d\n", V);
					  fprintf(fout,"timeout = %d\n", timeout);
                      fprintf(fout,"-----\n");
                      fprintf(fout,"AverageDelay = %f\n", delay);
					  fprintf(fout,"PacketGenerated = %f\n", pcktscount);
                      fprintf(fout,"PacketReceived = %f\n", receivedcount);
					  fprintf(fout,"Qsumlastslot = %d\n", qsum);
                      fprintf(fout,"sumTransmissions = %f\n", sum_num_transmissions);
                      fprintf(fout,"============================================\n");
                      printf("*");
                      fflush(fout);
                      if (V == 0)
                         V = 1;
                      else
                          V *=10;
                }
                printf("\n");                
				SQRTCX++;
          }
          printf("\n");
          timeout *= 2;
    }
    printf("\n");
    system("PAUSE");
    return EXIT_SUCCESS;
}
