#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
using namespace std;

#define MaxN 50
#define SIMTIME  1000000
#define QUEUESIZE 500 // used to create arrays
#define MaxSQRTC 6



typedef struct queuestruct {
    double q[QUEUESIZE][3];
    int si; // start index
    int count; // count
}queuetype;


int V;
int N;
int SQRTCX;
int SQRTCY;
int x[MaxN];
int y[MaxN];
double cell[MaxSQRTC][MaxSQRTC][6];
int U[MaxN][MaxN];
double timer[MaxN][MaxN];
double L; // number of copies
double t;
FILE *fout;
queuetype quar[MaxN][MaxN][2];
double delay;
double delaysum;
double difdelay;
int qsum;

double rand_p;
double lambda;
double sum_num_transmissions;
double receivedcount;
double pcktscount;
double seq_num[MaxN];
double seq_rcvd[MaxN];
int pcktrcvd[500000];
int maxdupQsize;
double dupQempty;
double bothQempty;
int initializeq(queuetype *qp)
{
    qp->si = 0;
    qp->count = 0;
    return (0);
}

int pushq(queuetype *qp, double v1, double v2, double v3)
{
    if (qp->count >= QUEUESIZE - 2) {      // if full
       fprintf(fout, "ERROR (push) queue is full\n");
       exit(1);
       
    }
    qp->q[(qp->si + qp->count) % QUEUESIZE][0] = v1;
    qp->q[(qp->si + qp->count) % QUEUESIZE][1] = v2;
	qp->q[(qp->si + qp->count) % QUEUESIZE][2] = v3;
    qp->count++;
    return(0);
}

double popq(queuetype *qp, double *v2, double *v3)
{
    if (qp->count <= 0) { //if empty
       fprintf(fout, "ERROR (pop) queue is empty\n");
       exit(1);
    }
    double v1;
    v1 = qp->q[qp->si][0];
    *v2 = qp->q[qp->si][1];
	*v3 = qp->q[qp->si][2];
    qp->si++;
    qp->si %=  QUEUESIZE;
    qp->count--;
    return(v1);    
}

double observeq(queuetype *qp, double *v2, double *v3) // read head without pop
{
    if (qp->count <= 0) { //if empty
		return(-1);
    }
    double v1;
    v1 = qp->q[qp->si][0];
    *v2 = qp->q[qp->si][1];
	*v3 = qp->q[qp->si][2];
    return(v1);    
}


double poplifoq(queuetype *qp, double *v2, double *v3)
{
    if (qp->count <= 0) { //if empty
       fprintf(fout, "ERROR (pop) queue is empty\n");
       exit(1);
    }
    double v1;
	
    v1 = qp->q[(qp->si + qp->count - 1) % QUEUESIZE][0];
    *v2 = qp->q[(qp->si + qp->count - 1) % QUEUESIZE][1];
	*v3 = qp->q[(qp->si + qp->count - 1) % QUEUESIZE][2];
    qp->count--;
    return(v1);    
}


int isemptyq(queuetype *qp)
{
    if (qp->count <= 0) { //if empty
       return(1);
    }
    return (0);
}

int isfullq(queuetype *qp)
{
    if (qp->count >= QUEUESIZE - 2) {      // if full
       return(1);
    }
    return (0);
}



int uniform_rnd(int n)
{
	int r;
	r = (int) (((double)rand() / ((double)(RAND_MAX) + (double)(1))) * n);
	return (r);
}


/* return uniform random variable in [0,1) */
double uniform_rnd_float()
{
	return(  (double)rand()  / (    (double)(RAND_MAX) + (double)(1)   ) );
}

void init_pos()
{
	int i;
	for(i = 0; i < N; i++){
		x[i] = uniform_rnd(SQRTCX);
		y[i] = uniform_rnd(SQRTCY);
	}
}

int is_received(double v1)
{
    int i;
    int j;
    int n;
    i = int(v1) / (int) 16;
    j = int(v1) % (int) 16;
    n = 1;
    n = n << j;
    n = n & pcktrcvd[i];
    if(n == 0) {
         return(0);
    }
    return(1);
}

int packet_received(double v1, double v2)
{
    int i;
    int j;
    int n;
	// v1 packet seqnum
	// v2 packet start time
	
	// if not already received
	if(!(is_received(v1))) {
    
         // update average delay
	     delaysum += t - v2;
	     receivedcount++;
	
	     // register this packet as already received
	     i = int(v1) / (int) 16;
	     j = int(v1) % (int) 16;
	     n = 1;
	     n = n << j;
	     pcktrcvd[i] = pcktrcvd[i] | n;
    }
     
	// else 
	// do nothing
	return(0);
}

int SIM()
{
	/*
	 * set to zero
	 */
	
	delaysum = 0;
    delay = 0;
	
	sum_num_transmissions = 0;
    receivedcount = 0;
	pcktscount = 0;
	int i;
	int ii;
	int j;
	int k;
	int tr;
	int rc;
	int cm;
	int cp;
	int cpflag;
	double v1; // packet seqnum
	double v2; // packet start time
	double v3; // num of copies
	double v1temp; // packet seqnum
	double v2temp; // packet start time
	double v3temp; // num of copies
	
	double wabc;
	double wabct;
	double timerabc;
	double threshold;
	
	double tiecount;
	
	for(i = 0; i < N; i++){
        for(j = 0; j < N; j++){  
            initializeq(&(quar[i][j][0]));
			initializeq(&(quar[i][j][1]));
            U[i][j] = 0;
			if (i!=j) {
				timer[i][j] = SIMTIME; //infinity
			}
			else{
				// i = j
				timer[i][j] = 0; //infinity
			}
        }
	}
	for(i = 0; i < 500000; i++) {
          pcktrcvd[i] = 0;
    }
	init_pos();
	t = 0;

	for(i = 0; i < N; i++) {
          seq_num[i] = 0;
          seq_rcvd[i] = 0;
    }
    maxdupQsize = 0;
	dupQempty = 0;
	bothQempty = 0;    
	while (t < SIMTIME) {
		init_pos();
		
		//duplicates removal
		for(i = 0; i < N; i++) {
			  for(j = 0; j < N; j++) {
				  if ( i !=j) {
					    for(ii = 0; ii < quar[i][j][0].count; ii++) {
								U[i][j]--;
								v1temp = popq(&(quar[i][j][0]), &v2temp, &v3temp);
								if(!is_received(j * lambda * 1.2 * SIMTIME + v1temp)){
									pushq(&(quar[i][j][0]),v1temp, v2temp, v3temp);
									U[i][j]++;
								}
						}
				  }
			  }
		}
		
		//initialize cell
		for(i = 0; i < SQRTCX; i++) {
              for(j = 0; j < SQRTCY; j++) {
					cell[i][j][4] = 0;
					cell[i][j][5] = 0;
					cell[i][j][0] = -1;
					cell[i][j][1] = -2;
					cell[i][j][2] = -3;
					cell[i][j][3] = 0;  //tiecount
              }
        }
		// scheduling
		
		for(i = 0; i < N; i++) {
              for(j = 0; j < N; j++) {
                    if (i != j) {
						  if(quar[i][j][1].count > maxdupQsize) {
								maxdupQsize = quar[i][k][1].count;
						  }
						  if(quar[i][j][1].count == 0){
								dupQempty++;
								if(quar[i][j][0].count == 0) {
									bothQempty++;
								}
						  }

                          if((x[i] == x[j]) && (y[i] == y[j])) {

							    // node i & j in the same cell 
                                 for(k = 0; k < N; k++){
                                       // for each comodity k
                                       wabc =  quar[i][k][0].count - quar[j][k][0].count; 
									   wabct = U[i][k] - U[j][k]; //quar[j][k][0].count; // -+ 0.005 * quar[j][k][1].count ;
									   timerabc = timer[i][k] - timer[j][k]; 
									   if((U[i][k] == 0) || (timerabc < 0) ){
										   // do nothing
									   }
									   else {
										   		if (j == k) {
												   if(cell[x[i]][y[i]][1] != cell[x[i]][y[i]][2]) {
														tiecount = 1.0;
													    cell[x[i]][y[i]][4] = wabc;
														cell[x[i]][y[i]][5] = timerabc;
														cell[x[i]][y[i]][0] = i;
														cell[x[i]][y[i]][1] = j;
														cell[x[i]][y[i]][2] = k;
														cell[x[i]][y[i]][3] = tiecount;
												   }
												   else {
													    tiecount = cell[x[i]][y[i]][3];
													    tiecount++;
														rand_p = uniform_rnd_float();
														if(rand_p < 1.0 / tiecount) {
															cell[x[i]][y[i]][4] = wabc;
															cell[x[i]][y[i]][5] = timerabc;
															cell[x[i]][y[i]][0] = i;
															cell[x[i]][y[i]][1] = j;
															cell[x[i]][y[i]][2] = k;
															cell[x[i]][y[i]][3] = tiecount;
														}
												   }
												}
										   		else  {
											   		 if(cell[x[i]][y[i]][1] != cell[x[i]][y[i]][2]){
														if (timerabc > cell[x[i]][y[i]][5]){
															tiecount = 1.0;
															cell[x[i]][y[i]][4] = wabc;
															cell[x[i]][y[i]][5] = timerabc;
															cell[x[i]][y[i]][0] = i;
															cell[x[i]][y[i]][1] = j;
															cell[x[i]][y[i]][2] = k;
															cell[x[i]][y[i]][3] = tiecount;
														}
														else if (timerabc == cell[x[i]][y[i]][5]){
															tiecount = cell[x[i]][y[i]][3];
															tiecount++;
															rand_p = uniform_rnd_float();
															if(rand_p < 1.0 / tiecount) {
																cell[x[i]][y[i]][4] = wabc;
																cell[x[i]][y[i]][5] = timerabc;
																cell[x[i]][y[i]][0] = i;
																cell[x[i]][y[i]][1] = j;
																cell[x[i]][y[i]][2] = k;
																cell[x[i]][y[i]][3] = tiecount;
															}
														}
											   		}
										   		}
									   }
                                 }
                          }
                    }
              }
        }
        // updating queues
		for(i = 0; i < SQRTCX; i++) {
              for(j = 0; j < SQRTCY; j++) {
                    if ( (cell[i][j][4] > 0) || (cell[i][j][5] > 0)) {
                         //we have packet transmition 
                         // from node cell[i][j][0]
                         tr = cell[i][j][0];
                         // to node cell[i][j][1]
                         rc = cell[i][j][1];
                         // comodity cell[i][j][2]
                         cm = cell[i][j][2];

                         
						 if(!(isemptyq(&(quar[tr][cm][0])))) {
							v1 = popq(&(quar[tr][cm][0]), &v2, &v3);
							U[tr][cm]--;
							cpflag = 0;  // indication that this packet from original queue
						 }
						 else {
							printf("error: no packets to transmit\n");
							// all code here should not be executed in spray and focus case
							v1 = popq(&(quar[tr][cm][1]), &v2, &v3); // poplifoq
							U[tr][cm]--;
							cpflag = 1;  // indication that this packet from duplicates queue
						 }
						 if(rc == cm){
                             // packet received
							 sum_num_transmissions++;
                             packet_received(cm * lambda * 1.2 * SIMTIME + v1, v2);
                             if(v1 == seq_rcvd[cm]){
                                   seq_rcvd[cm]++;
                                   while(is_received(cm * lambda * 1.2 * SIMTIME + seq_rcvd[cm])) {
                                       seq_rcvd[cm]++;    
                                   }        
                             }
                         }
						 else {
							  if (v3 == 1) {
								    // for spray and wait we do nothing but we need to return packet in the sender
									// sum_num_transmissions++;
									pushq(&(quar[tr][cm][0]),v1, v2, v3);
									U[tr][cm]++;                              
							  }
							  else if (v3 > 1){
								  double v3_1;
								  double v3_2;
								  v3_1 =  (int) v3 / 2;
								  v3_2 = v3 - v3_1;
								  sum_num_transmissions++;
								  pushq(&(quar[tr][cm][0]),v1, v2, v3_1);
								  U[tr][cm]++;                              
								  pushq(&(quar[rc][cm][0]),v1, v2, v3_2);
								  U[rc][cm]++;                              
							  }
							  else {
								  printf("error: number of copies must be more than 1");
							  }
                         }
					}
              }
        }
		//computing timers
		for(i = 0; i < N; i++) {
			for(j = 0; j < N; j++) {
				if (i != j) {
					timer[i][j]++;
				} 
				else{
					// i = j
					timer[i][j] = 0;
				}
			}
		}
		for(i = 0; i < N; i++) {
			for(j = i+1; j < N; j++) {
					if((x[i] == x[j]) && (y[i] == y[j])) {
						timer[i][j] = 0;
						timer[j][i] = 0;
						int d;
						for(d = 0; d < N; d++) {
							if((timer[j][d] + 1) < timer[i][d]) {
								timer[i][d] = timer[j][d] + 1;
							}
							else if((timer[i][d] + 1) < timer[j][d]) {
								timer[j][d] = timer[i][d] + 1;
							}
						}
					}
			}
		}

        
        // arrivals 
		for(i = 0; i < N; i++) {
              rand_p = uniform_rnd_float();
			  if(rand_p < lambda) {
				  // we have an arrival
				  pcktscount++;
				  if (i % 2 == 0) {
					  j = i + 1;
				  }
				  else {
					  j = i - 1;
				  }
                  U[i][j]++;
                  pushq(&(quar[i][j][0]),seq_num[j], t, L);
                  (seq_num[j])++;
              }        
        }
		t++;
	}
	qsum = 0;
	for(i = 0; i < N; i++) {
		for(j = 0; j < N; j++) {
			qsum += U[i][j];
		}
	}


	difdelay = delay - (delaysum / receivedcount);
    delay = delaysum / receivedcount;
	return(0);
}





int main(int argc, char *argv[])
{
    srand((unsigned)time(NULL));
    fout = fopen("out.txt","wt");
    lambda = 0.08;
	L = 5;
    while (lambda < 0.085) {
	//while (L <= 31) {
		  SQRTCX = 5;
          while (  SQRTCX <= 5) {
			    SQRTCY = SQRTCX;
			    N = (int) SQRTCX * SQRTCY * 1.7399;
				if (N % 2 == 1) {
					N++;
				}
                V = 1;
                while (V <= 1) {
                      SIM();
                      fprintf(fout,"Spray and Focus L = %f\n", L);
					  fprintf(fout,"lambda = %f\n", lambda);
                      fprintf(fout,"N = %d\n", N);
                      fprintf(fout,"V = %d\n", V);
                      fprintf(fout,"-----\n");
                      fprintf(fout,"AverageDelay = %f\n", delay);
					  fprintf(fout,"PacketGenerated = %f\n", pcktscount);
                      fprintf(fout,"PacketReceived = %f\n", receivedcount);
					  fprintf(fout,"Qsumlastslot = %d\n", qsum);
                      fprintf(fout,"sumTransmissions = %f\n", sum_num_transmissions);
                      fprintf(fout,"============================================\n");
                      printf("*");
                      fflush(fout);
                      if (V == 0)
                         V = 1;
                      else
                          V *=10;
                }
                printf("\n");                
				SQRTCX++;
          }
          printf("\n");
          lambda *= 2;
		  //L += 5;
    }
    printf("\n");
    system("PAUSE");
    return EXIT_SUCCESS;
}
